@@ -24,6 +24,8 @@ import shutil
 from datetime import datetime
 from io import RawIOBase
 
+import botocore
+
 from barman.clients.cloud_compression import decompress_to_file
 from barman.cloud import (
     DEFAULT_DELIMITER,
@@ -41,6 +43,9 @@ from barman.exceptions import (
     SnapshotInstanceNotFoundException,
 )
 
+_logger = logging.getLogger(__name__)
+
+
 try:
     # Python 3.x
     from urllib.parse import urlencode, urlparse
@@ -201,7 +206,7 @@ class S3CloudInterface(CloudInterface):
             self.bucket_exists = self._check_bucket_existence()
             return True
         except EndpointConnectionError as exc:
-            logging.error("Can't connect to cloud provider: %s", exc)
+            _logger.error("Can't connect to cloud provider: %s", exc)
             return False
 
     def _check_bucket_existence(self):
@@ -232,7 +237,7 @@ class S3CloudInterface(CloudInterface):
         # Get the current region from client.
         # Do not use session.region_name here because it may be None
         region = self.s3.meta.client.meta.region_name
-        logging.info(
+        _logger.info(
             "Bucket '%s' does not exist, creating it on region '%s'",
             self.bucket_name,
             region,
@@ -415,25 +420,102 @@ class S3CloudInterface(CloudInterface):
 
     def _delete_objects_batch(self, paths):
         """
-        Delete the objects at the specified paths
+        Deletes multiple objects from the S3 bucket at the specified paths.
 
-        :param List[str] paths:
+        Attempts to delete objects in batch using the `delete_objects()` API. If a
+        :exc:`ClientError` with the :exc:``MissingContentMD5` code is raised (common with older or
+        non-compliant object storage systems), falls back to deleting each object
+        individually using `delete_object()`.
+
+        Logs errors for any objects that fail to be deleted in bulk. Raises
+        :exc:`CloudProviderError` if any bulk deletion errors occur.
+
+        :param paths: List of object paths (keys) to delete from the S3 bucket.
+        :raises CloudProviderError:
+            If bulk deletion fails for any object.
+        :raises botocore.exceptions.ClientError:
+            If an AWS client error occurs that is not handled by the fallback logic.
         """
         super(S3CloudInterface, self)._delete_objects_batch(paths)
 
-        resp = self.s3.meta.client.delete_objects(
-            Bucket=self.bucket_name,
-            Delete={
-                "Objects": [{"Key": path} for path in paths],
-                "Quiet": True,
-            },
-        )
-        if "Errors" in resp:
-            for error_dict in resp["Errors"]:
-                logging.error(
-                    'Deletion of object %s failed with error code: "%s", message: "%s"'
-                    % (error_dict["Key"], error_dict["Code"], error_dict["Message"])
+        try:
+            resp = self.s3.meta.client.delete_objects(
+                Bucket=self.bucket_name,
+                Delete={
+                    "Objects": [{"Key": path} for path in paths],
+                    "Quiet": True,
+                },
+            )
+            if "Errors" in resp:
+                for error_dict in resp["Errors"]:
+                    _logger.error(
+                        'Bulk deletion of object %s failed with error code: "%s", '
+                        'message: "%s"'
+                        % (
+                            error_dict["Key"],
+                            error_dict["Code"],
+                            error_dict["Message"],
+                        )
+                    )
+                raise CloudProviderError()
+        except botocore.exceptions.ClientError as e:
+            # Fallback for MissingContentMD5 errors: `_delete_object` deletes a single
+            # object at a time using `.delete()` for the given key. This is necessary
+            # because bulk delete may fail when Content-MD5 headers are missing.
+            # Note: The S3 protocol returns a 'MissingContentMD5' error code in that
+            # situation. Some "S3-compatible" storage systems are not fully compatible,
+            # and return "InvalidRequest" instead, so we also check the message in that.
+            if (
+                e.response["Error"]["Code"] == "MissingContentMD5"
+                or "missing required header for this request: content-md5"
+                in e.response["Error"]["Message"].lower()
+            ):
+                for path in paths:
+                    self._delete_object(path)
+            else:
+                raise e
+
+    def _delete_object(self, path):
+        """
+        Delete an object from AWS S3 and use it as a fallback method.
+
+        Attempts to delete an object individually. If a deletion fails due to a client
+        error, logs the error details and raises a :exc:`CloudProviderError`.
+        Any other exceptions are also logged and re-raised as :exc:`CloudProviderError`.
+
+        .. note::
+            This method should not be called as the "primary deletion" method. Its use
+            is strictly for fallback purposes, when users are using older or
+            non-compliant S3-like object storage systems.
+
+        :param str path: S3 object key (path) to delete.
+        :type path: str
+        :raises CloudProviderError: If deletion of any object fails.
+        """
+        try:
+            _logger.warning(
+                "Bulk delete failed with 'MissingContentMD5'. Falling back "
+                "to deleting this file individually: %s." % path
+            )
+            self.s3.meta.client.delete_object(
+                Bucket=self.bucket_name,
+                Key=path,
+            )
+        except botocore.exceptions.ClientError as e:
+            error_code = e.response["Error"]["Code"]
+            error_message = e.response["Error"]["Message"]
+            _logger.error(
+                "Deletion of object %s failed with error code: "
+                '"%s", message: "%s"'
+                % (
+                    path,
+                    error_code,
+                    error_message,
                 )
+            )
+            raise CloudProviderError()
+        except Exception as e:
+            _logger.error("Deletion of object %s failed with error: %s" % (path, e))
             raise CloudProviderError()
 
     def get_prefixes(self, prefix):
@@ -460,15 +542,37 @@ class S3CloudInterface(CloudInterface):
             raise ValueError(
                 "Deleting all objects under prefix %s is not allowed" % prefix
             )
+
         bucket = self.s3.Bucket(self.bucket_name)
-        for resp in bucket.objects.filter(Prefix=prefix).delete():
-            response_metadata = resp["ResponseMetadata"]
-            if response_metadata["HTTPStatusCode"] != 200:
-                logging.error(
-                    'Deletion of objects under %s failed with error code: "%s"'
-                    % (prefix, response_metadata["HTTPStatusCode"])
-                )
-                raise CloudProviderError()
+        try:
+            # Although this is written as a `for` loop, the `.delete()` call on the
+            # filtered objects performs a **bulk delete** in AWS S3 in a single API
+            # request. Each `resp` is the metadata for a deleted object returned by the
+            # bulk delete operation.
+            for resp in bucket.objects.filter(Prefix=prefix).delete():
+                response_metadata = resp["ResponseMetadata"]
+                if response_metadata["HTTPStatusCode"] != 200:
+                    _logger.error(
+                        'Deletion of objects under %s failed with error code: "%s"'
+                        % (prefix, response_metadata["HTTPStatusCode"])
+                    )
+                    raise CloudProviderError()
+        except botocore.exceptions.ClientError as e:
+            # Fallback for MissingContentMD5 errors: `_delete_object` deletes a single
+            # object at a time using `.delete()` for the given key. This is necessary
+            # because bulk delete may fail when Content-MD5 headers are missing.
+            # Note: The S3 protocol returns a 'MissingContentMD5' error code in that
+            # situation. Some "S3-compatible" storage systems are not fully compatible,
+            # and return "InvalidRequest" instead, so we also check the message in that.
+            if (
+                e.response["Error"]["Code"] == "MissingContentMD5"
+                or "missing required header for this request: content-md5"
+                in e.response["Error"]["Message"].lower()
+            ):
+                for obj in bucket.objects.filter(Prefix=prefix):
+                    self._delete_object(obj.key)
+            else:
+                raise e
 
 
 class AwsCloudSnapshotInterface(CloudSnapshotInterface):
@@ -743,7 +847,7 @@ class AwsCloudSnapshotInterface(CloudSnapshotInterface):
             volume_name,
             backup_info.backup_id.lower(),
         )
-        logging.info(
+        _logger.info(
             "Taking snapshot '%s' of disk '%s' (%s)",
             snapshot_name,
             volume_name,
@@ -832,7 +936,7 @@ class AwsCloudSnapshotInterface(CloudSnapshotInterface):
         # successful state. If the successful state is not reached after the maximum
         # number of attempts (default: 40) then a WaiterError is raised.
         snapshot_ids = [snapshot.identifier for snapshot in snapshots]
-        logging.info("Waiting for completion of snapshots: %s", ", ".join(snapshot_ids))
+        _logger.info("Waiting for completion of snapshots: %s", ", ".join(snapshot_ids))
         waiter = self.ec2_client.get_waiter("snapshot_completed")
         waiter.wait(
             SnapshotIds=snapshot_ids,
@@ -874,7 +978,7 @@ class AwsCloudSnapshotInterface(CloudSnapshotInterface):
                     value = value.isoformat()
                 _output[key] = value
 
-        logging.info("Snapshot locked: \n%s" % json.dumps(_output, indent=4))
+        _logger.info("Snapshot locked: \n%s" % json.dumps(_output, indent=4))
 
     def _delete_snapshot(self, snapshot_id):
         """
@@ -889,7 +993,7 @@ class AwsCloudSnapshotInterface(CloudSnapshotInterface):
             # If the snapshot could not be found then deletion is considered successful
             # otherwise we raise a CloudProviderError
             if error_code == "InvalidSnapshot.NotFound":
-                logging.warning("Snapshot {} could not be found".format(snapshot_id))
+                _logger.warning("Snapshot {} could not be found".format(snapshot_id))
             elif error_code == "SnapshotLocked":
                 raise SystemExit(
                     "Locked snapshot: %s.\n"
@@ -901,7 +1005,7 @@ class AwsCloudSnapshotInterface(CloudSnapshotInterface):
                     "Deletion of snapshot %s failed with error code %s: %s"
                     % (snapshot_id, error_code, exc.response["Error"])
                 )
-        logging.info("Snapshot %s deleted", snapshot_id)
+        _logger.info("Snapshot %s deleted", snapshot_id)
 
     def delete_snapshot_backup(self, backup_info):
         """
@@ -910,7 +1014,7 @@ class AwsCloudSnapshotInterface(CloudSnapshotInterface):
         :param barman.infofile.LocalBackupInfo backup_info: Backup information.
         """
         for snapshot in backup_info.snapshots_info.snapshots:
-            logging.info(
+            _logger.info(
                 "Deleting snapshot '%s' for backup %s",
                 snapshot.identifier,
                 backup_info.backup_id,
